//##################################################################################
//
//This is the software for the motor inverter of the BASTLI circular saw.
//
//Circular saw built and programmed by Mario Mauerer.
//Power supply of the saw built and programmed by Lukas Schrittwieser.
//www.bastli.ethz.ch
//April 2011
//
//##################################################################################

//##### INCLUDES #####
#include <avr/io.h>        
#include <avr/interrupt.h>
#include <util/delay.h>

#include "main_saw.h" //Headerfile for main.c
#include "uart_saw.h" //Uart-Functions
#include "startup_saw.h" //startup the motor
//#################

//#### Global Variables: ####
volatile uint8_t state = 0; //state of the commutation, 0-5
volatile uint8_t startupflag = 1; //says whether the motor is spinning up or running normally. 
volatile uint8_t zcflag = 0; //flag that allows the ZC-Measurement afther the commutation-holdoff-time has elapsed. 

volatile	uint16_t zctime = 0; //measures the time between commutation and ZC. 
volatile	uint16_t zctimeold = 50; //variable for the commutation-timing-IIR-Filter
volatile  	uint16_t zctimenew = 0; //variable for the commutation-timing-IIR-Filter         //this variable could be deleted and replaced

volatile uint8_t zcdetectionmask = 0; //mask for measuring the ZC ==> on which Portpin has to be measured?
volatile uint8_t zcdetectioninvmask = 0; //mask for measuring the ZC ==> Does the Portbin have to be high or low?

//for the uart:
volatile char *uartpoint; //this points to the current byte that is sent via the uart.
volatile uint8_t uartstat = 0; //this represents the status of the uart-transmission.

//for the controller:
volatile uint8_t t2ovfcnt = 0;
volatile uint8_t cntrlstate = 0;

volatile uint16_t commcnt = 100;
volatile uint16_t accu = 200;
volatile uint16_t pwm = 0;


volatile uint16_t rotsoll = 230; //standard-value: 5000RPM = 230

//for the ADC:
volatile uint8_t adccnt = 0;
volatile uint16_t adcvalue = 0;

//for shutting the saw down:
volatile uint8_t shutdownflag = 0;
volatile uint8_t measurecnt = 0;

//for watchdog:
volatile uint8_t wdtovf = 0;

//for boost-button:
volatile uint8_t boostcnt = 0;
//##################

//##### MAIN-ROUTINE #####
int main(void)
{	
	port_init(); //Initialize the IO-ports of the AVR
	peripheral_init(); //Initialize the peripheral Stuff of the AVR (except timers)
	system_check(); //perform a system-check prior to the start
	motor_beep(); //adjust the motor position and let it beep three times
	timer_init(); //init all timers and thus start up the saw. 

	uint16_t help = 0;
	float temp = 0;
	uint16_t temperature = 0;
	uint16_t current = 0;
	uint16_t voltage = 0;
	
	uint8_t boostflag = 0;
	
	while(1)
	{
		//boost-button: //Don't use without a break-chopper in the system!!
		if(!(PINB & (1<<PB5))) //if button pressed
		{
			_delay_ms(20); 
			boostflag = 1;
			boostcnt = 0;
			rotsoll = 202;  //11000RPM
		}
		if(boostflag == 1 && boostcnt >= 2)
		{
			boostflag = 0;
		}
		//######################
		
		
		//#####Send data to terminal:###### //NOT USABLE: NOT ENOUGH FLASH
		/*if(wdtovf == 1) //transmit some system parameters to uart!
		{
			/*float help = 0;
			uint16_t help2 = 0;
			uint16_t help3 = 0;
			
			uart_stringsend("RPM"); 
			help = 54.496 * rotsoll; //change this value when changing the controller-execution-time! RPM = 54.496*rotsoll
			help2 = (uint16_t)help;
			_delay_ms(20);
			uart_intsend(help2);
			_delay_ms(10);
			
			/*help2 = 45*voltage+47;
			help2 = help2/1000;
			uart_stringsend("DC-Link Voltage (V):");
			_delay_ms(20);
			uart_intsend(help2);
			_delay_ms(5);
			
			help2 = 18*current+1880;
			help2 = help2/1000;
			uart_stringsend("Current (A):");
			_delay_ms(20);
			uart_intsend(help2);
			_delay_ms(5);
			
			help2 =44*temperature-18800;
			help2 = help2/100;
			help3 = 2*temperature;
			help3 = help3/1000;
			help2 = help2+help3;
			
			uart_stringsend("Temperature (C):");
			_delay_ms(20);
			uart_intsend(help2);

			wdtovf = 0;
		}*/
		//###############################
		
		//### Evaluate the Speed-Poti and control the rotation-frequency: ###
		//Value: 55<rotsoll<202 ==> ca.3000rpm<rotsoll<ca. 11000rpm (202 only for boost-button!)
		if(boostflag == 0)
		{
			ADCSRA &= ~(1<<ADSC); //ADC Disable in order to change channels.
			ADMUX &= 0b11110000; //Clear current Channel Selection
			ADMUX |= ((1<<MUX0) | (1<<MUX1)); //ADC3 as Input (Speed-Poti)
			ADCSRA |= (1<<ADSC); //restart ADC
			ADCSRA |= (1<<ADIE); //enable the conversion complete Interrupt
			adccnt = 0;
			adcvalue = 0;
			while(adccnt < 16) //take 16 measurements (using the ADC-Interrupt)
			{
			}
			ADCSRA &= ~(1<<ADIE); //disable the conversion complete Interrupt
			help = adcvalue >> 4; //divide by 16 to calculate the average
			temp = help * 0.12; //do some black magic corrections to get the value into the desired range :-) 
			help = (uint16_t)temp;
			help = help + 55;
			if(help >= 170)
				help = 170;
			if(help <= 55)
				help = 55;
			cli();
			rotsoll = help;
			sei();
		}
		//#########################################
		
		//### Measure Temperature of the Motor: Should be OK: ###
		//Temperature (C) = 0.442 * ADC - 188
		//configure ADC for Port ADC2 (Temp-Sensor)
		ADCSRA &= ~(1<<ADSC); //ADC Disable in order to change channels.
		ADMUX &= 0b11110000; //Clear current Channel Selection
		ADMUX |= (1<<MUX1); //ADC2 as Input (Temp-Sensor)
		ADCSRA |= (1<<ADSC); //restart ADC
		ADCSRA |= (1<<ADIE); //enable the conversion complete Interrupt
		adccnt = 0;
		adcvalue = 0;
		while(adccnt < 16) //take 16 measurements (using the ADC-Interrupt)
		{
		}
		ADCSRA &= ~(1<<ADIE); //disable the conversion complete Interrupt
		help = adcvalue >> 4; //divide by 16 to build average
		temperature = help;
		if(help > 495) //if temp > 31C ==> start the Fan!
		{
			FAN_on;
		}
		else if(help < 486) //if temp < 27C ==> shut the Fan off
		{
			FAN_off;
		}
		if(help > 545) //if temp > 53C ==> shutdown!
		{
			scram(1);
		}
		//###################################
		
		//### Measure the DC-Link and check for shutdown: ###
		//DC-Link-Voltage = 0.045 * ADC + 0.047
		//configure ADC for Port ADC1 : DC-Link-Voltage:
		ADCSRA &= ~(1<<ADSC); //ADC Disable in order to change channels.
		ADMUX &= 0b11110000; //Clear current Channel Selection
		ADMUX |= (1<<MUX0); //ADC1 as Input (DC-Link-Voltage)
		ADCSRA |= (1<<ADSC); //restart ADC
		ADCSRA |= (1<<ADIE); //enable the conversion complete Interrupt
		adccnt = 0;
		adcvalue = 0;
		while(adccnt < 16) //take 16 measurements (using the ADC-Interrupt)
		{
		}
		ADCSRA &= ~(1<<ADIE); //disable the conversion complete Interrupt
		help = adcvalue >> 4; //divide by 16 to build average
		voltage = help;
		if((help < 443 || help > 1000) && measurecnt >= 50) //if DC-Link-Voltage larger than 43V or smaller than 15V ==> not good ==> power the saw down in a quick ramp! (no sudden stop as that would be too hard for the mechanics)
		{
			scram(2);
		}
		//#####################################
		
		//### Measure Current:  ###
		//Current = 0.0181 * ADC + 1.88
		//configure ADC for Port ADC0: Current
		ADCSRA &= ~(1<<ADSC); //ADC Disable in order to change channels.
		ADMUX &= 0b11110000; //ADC0 as Input (Current-Measurement)
		ADCSRA |= (1<<ADSC); //restart ADC
		ADCSRA |= (1<<ADIE); //enable the conversion complete Interrupt
		adccnt = 0;
		adcvalue = 0;
		while(adccnt < 16) //take 16 measurements (using the ADC-Interrupt)
		{
		}
		ADCSRA &= ~(1<<ADIE); //disable the conversion complete Interrupt
		help = adcvalue >> 4; //divide by 16 to build average
		current = help;
		if(help > 1022) //too much current (>ca. 18A) 
		{
			scram(3);
		}
		//###################################
	
	}
	
	return 0;
}
//#####################

//#####SYSTEM-INIT#########
void port_init(void)
{
//OUTPUTS:
	DDRD |= ((1<<PD2) | (1<<PD3) | (1<<PD4) | (1<<PD5) | (1<<PD6) | (1<<PD7)); // Outputs for the Gatedrives
	
	DDRC |= (1<<PC4); //LED-Output
	
	DDRC  |= (1<<PC5); //Fan-enable-Output
	
//INPUTS:
	DDRB &= ~((1<<PB0) | (1<<PB1) | (1<<PB2)); // Comparator-Inputs
	
	DDRB &= ~(1<<PB5); //"Boost"-Button Input
	PORTB |= (1<<PB5); //enable pullup for the "Boost"-Button 
	
	DDRC &= ~(1<<PC2); //Temp-Sensor Input; ADC2
	
	DDRC &= ~(1<<PC1); //DC link voltage measurement input; ADC1
	
	DDRC &= ~(1<<PC0); //Current measurement input; ADC0
	
	DDRC &= ~(1<<PC3); //"Speed"-Poti input; ADC3
	
	//Shut everything off:
	LED_off;
	FAN_off;
	AHIGH_off;
	ALOW_off;
	BHIGH_off;
	BLOW_off;
	CHIGH_off;
	CLOW_off;
}

void peripheral_init(void)
{
//UART: 9600Baud, Asynchronous, no Parity, 1 Stop-bit, 8-bit data
	UCSR0B |= ((1<<RXEN0) | (1<<TXEN0)); //Enable Receiver and Transmitter
	UCSR0C |= ((1<<UCSZ01) | (1<<UCSZ00)); //8-Bit
	UBRR0 = 129; //Baud-Rate: 9600
	
//ADC:
	ADCSRA |= ((1<<ADEN) | (1<<ADATE) | (ADPS0) | (1<<ADPS1) | (1<<ADPS2)); //ADC enable, free running, prescaler = 128
	ADMUX |= (1<<REFS0); //AVCC as Reference.
	ADMUX |= ((1<<MUX0) | (1<<MUX1)); //ADC3 as Input (Speed-Poti)
	ADCSRA |= (1<<ADSC);	//Start ADC
	
	SREG |= (1<<7); //enable interrupts globally
}

void timer_init(void)
{
//Watchdog-Timer: For putting the system Params to the UART
	cli();
	WDTCSR |= ((1<<WDCE)  | (1<<WDE)); //let prescaler values change
	WDTCSR = 0b10100000; //prescaler: int-period = 4s, no system reset
	WDTCSR |= (1<<WDIE); //enable interrupt
	sei();
	
//TIMER1: PWM
	TCCR1A |= (1<<WGM11); //Fast-PWM, 9-Bit, Prescaler = 1 ==> 40kHz
	TCCR1B |= ((1<<WGM12) | (1<<CS10));
	OCR1A = 215; //D at beginning
	TIMSK1 |= ((1<<OCIE1A) | (1<<TOIE1));
	
//TIMER0: commutation timing:
	TCCR0A |= (1<<WGM01); //CTC on OCR1A
	TCCR0B |= ((1<<CS01) | (1<<CS00)); //p = 64
	OCR0A |=  180;
	OCR0B |= 15;
	TIMSK0 |= (1<<OCIE0A); 
	
//TIMER2: controller-Timing
	TCCR2B |= ((1<<CS20) | (1<<CS21) | (1<<CS22)); //prescaler = 1024, start timer
	TIMSK2 |= (1<<TOIE2); //enable OVF-Interrupt
	TIMSK2 |= (1<<OCIE2A); //enable Compare-Interrupt 2
	OCR2A = 50; //random value
}
void scram(uint8_t param)
{
	cntrlstate = 0; //no more control
	shutdownflag = 1; //ramp it down!
	while(shutdownflag == 1) //wait for rampdown to finish
	{
	}
	
	//shut everything off:
	cli(); //no more interrupts from the timers
	ALOW_on; //block the motor:
	BLOW_on;
	CLOW_on;
	
	if(param == 1) //Temp shutdown
	{
		while(1) //let the LED blink THREE times when having temperature-error and wait forever here.
		{
			LED_on;
			_delay_ms(200);
			LED_off;
			_delay_ms(200);
			LED_on;
			_delay_ms(200);
			LED_off;
			_delay_ms(200);
			LED_on;
			_delay_ms(200);
			LED_off;
			_delay_ms(600);
		}	
	}
	else if(param == 2) //DC-Voltage
	{
		while(1) //let the LED blink TWO times when having a DC-Link-Error and wait forever here. 
		{
			LED_on;
			_delay_ms(100);
			LED_off;
			_delay_ms(100);
			LED_on;
			_delay_ms(100);
			LED_off;
			_delay_ms(500);
		}
	}
	else if(param == 3)//Current
	{
		while(1) //let the LED blink FOUR times when having a current-error and wait here forever
		{
			LED_on;
			_delay_ms(200);
			LED_off;
			_delay_ms(200);
			LED_on;
			_delay_ms(200);
			LED_off;
			_delay_ms(200);
			LED_on;
			_delay_ms(200);
			LED_off;
			_delay_ms(200);
			LED_on;
			_delay_ms(200);
			LED_off;
			_delay_ms(600);
		}
			
	}
	
		
	
}

//#####################

//########### ISR ############
ISR(TIMER1_COMPA_vect ) //Timer0-Compare: PWM //Shut the normal output off and enable the low-side FET ==> Freewheeling
{
	switch(state)
	{
		case 0:
			//PORTD = 0b01000000; //active freewheeling
			PORTD = 0b01000100; //BL:ON, AH: OFF
		break;

		case 1:
			//PORTD = 0b00010000; //active freewheeling
			PORTD = 0b00010100; //CL:ON, AH: OFF
		break;

		case 2:
			//PORTD = 0b00000100; //active freewheeling
			PORTD = 0b00010100; //CL: ON, BH: OFF
		break;

		case 3:
			//PORTD = 0b01000000; //active freewheeling
			PORTD = 0b01010000; //AL: ON, BH: OFF
		break;

		case 4:
			//PORTD = 0b00010000; //active freewheeling
			PORTD = 0b01010000; //AL: ON, CH: OFF
		break;

		case 5:
			//PORTD = 0b00000100; //active freewheeling
			PORTD = 0b01000100; //BL: ON, CH: OFF
		break;
	}	
}


ISR(TIMER1_OVF_vect ) //Timer1-OVF: PWM //turn the normal output on and disable the low-side FET.==> PWM-ON-Phase ==> Power! and ZC-Measurement
{
	
	//Check for Zero-Crossing!
	if(zcflag == 1) //pwm-off-period: measurement is allowed ==> measure, if there was a ZC. 
	{
		if((zcdetectionmask & (PINB ^ zcdetectioninvmask) )== 0) //commutation has happened!
		{
			zcflag = 0;
			zctime = TCNT0; //commutation time
			
			zctimenew = zctime + 3 * zctimeold; //IIR-Filter
			zctimenew= zctimenew / 4; //IIR-Filter
			TCNT0 = 0; //Reset the commutation timer
			OCR0A = zctimenew; //time to commutate
			OCR0B = zctimenew/8; //ZC holdoff-detection-fraction
			
			
			zctimeold = zctimenew; //IIR-Filter
			
			TIMSK0 |= (1<<OCIE0A); //OCR0A-interrupt enable ==> then: commutate!
		}
	}
	
	switch(state)
	{
		case 0:
			PORTD = 0b01001100; //AH, BL: ON
		break;

		case 1:
			PORTD = 0b00011100; //AH, CL: ON
		break;

		case 2:
			PORTD = 0b00110100; //BH,CL : ON
		break;

		case 3:
			PORTD = 0b01110000; //AL, BH: ON
		break;

		case 4:
			PORTD = 0b11010000; //AL, CH: ON
		break;

		case 5:
			PORTD = 0b11000100; //CH, BL: ON		
		break;
	}

}

ISR(TIMER0_COMPA_vect ) //Used for startup and commutation timing
{
	if(startupflag == 0) // if auto-commutation ==> commutate!
	{
		state++; //next commutation step
		if (state >= 6)
			state = 0;

		switch(state)
		{
			case 0:
				PORTD = 0b01001100; //AH, BL: ON
				zcdetectionmask = 0b00000001; //Detect on PB0 for 0
				zcdetectioninvmask = 0b0000000;
			break;

			case 1:
				PORTD = 0b00011100; //AH, CL: ON
				zcdetectionmask = 0b00000010; //Detect on PB1 for 1
				zcdetectioninvmask = 0b0000010;
			break;

			case 2:
				PORTD = 0b00110100; //BH,CL : ON
				zcdetectionmask = 0b00000100; //Detect on PB2 for 0
				zcdetectioninvmask = 0b0000000;
			break;

			case 3:
				PORTD = 0b01110000; //AL, BH: ON
				zcdetectionmask = 0b00000001; //Detect on PB0 for 1
				zcdetectioninvmask = 0b0000001;
			break;

			case 4:
				PORTD = 0b11010000; //AL, CH: ON
				zcdetectionmask = 0b00000010; //Detect on PB1 for 0
				zcdetectioninvmask = 0b0000000;
			break;

			case 5:
				PORTD = 0b11000100; //CH, BL: ON	
				zcdetectionmask = 0b00000100; //Detect on PB2 for 1
				zcdetectioninvmask = 0b0000100;
			break;
		}
		
		TCNT1 = 0; //reset PWM after commutation 

		TIMSK0 &= ~(1<<OCIE0A); //Disable the OCR1A-Interrupt
		TIMSK0 |= (1<<OCIE0B); //Enable the ZC-detection holdoff time interrupt
		commcnt++; //for the controller ==> count the commutations!
	}
	else if(startupflag == 1) //start the motor up
	{
		state++;
		if(state == 6)
			state = 0;

		switch(state)
		{
			case 0:
				PORTD = 0b01001100; //AH, BL: ON
			break;

			case 1:
				PORTD = 0b00011100; //AH, CL: ON
			break;

			case 2:
				PORTD = 0b00110100; //BH,CL : ON
			break;

			case 3:
				PORTD = 0b01110000; //AL, BH: ON
			break;

			case 4:
				PORTD = 0b11010000; //AL, CH: ON
			break;

			case 5:
				PORTD = 0b11000100; //CH, BL: ON		
			break;
			}

		TCNT1 = 0; //reset PWM

		OCR0A = OCR0A - 1 ; //increase the motor speed
		if(OCR0A <= 170) //if enough speed (OCR0A begins with 180)
		{
			startupflag = 0; //switch to auto-commutation.
			TIMSK0 &= ~(1<<OCIE0A); //Disable the OCR1A-Interrupt
			OCR1A =  100; //reduce the duty cyle for the rampup
			OCR0B = 15; 
			TIMSK0 |= (1<<OCIE0B); //Enable the ZC-detection holdoff time interrupt
		}
			
	}
	
}

ISR(TIMER0_COMPB_vect ) //used for ZC-detection-holdoff timing
{
	TIMSK0 &= ~(1<<OCIE0B); //Disable this ZC-detection holdoff time interrupt
	if(startupflag == 0) //only allow the ZC-measurement after the startup!
		zcflag = 1; //Enough time since the commutation has passed ==> allow the ZC-measurement!
}

ISR(USART_TX_vect ) //USART-TX-Complete
{
	if(*uartpoint) //if current transmission is not yet finished
	{
		UDR0 = *uartpoint; //send next byte
		uartpoint++; //next char.
	}
	else
	{
		UCSR0B &= ~(1<<TXCIE0);//disable TX-complete-Interrupt
		uartstat = 0; //finished with the transmission.
	}
}

ISR(TIMER2_OVF_vect ) //for the controller!
{
	measurecnt++; //delay for the dc-link measurement (may not measure at startup)
	if(measurecnt >= 250)
		measurecnt = 250;
	
	t2ovfcnt++;
	if(t2ovfcnt >= 2 && cntrlstate == 1) //execute the controller every 26ms
	{
		int16_t err = 0;

		t2ovfcnt = 0;
		
		err = rotsoll-commcnt; 	
		commcnt = 0;
		
		accu = accu + err; //accu, with anti-windup
		
		if(accu < 16) //limit the accu
			accu = 16;
		if(accu > 800)
			accu = 800;
		
		pwm = 12 * err + 10* accu; //PI-Controller
		
		pwm = pwm >> 4; //rescale the values for better parameter-possibilities above
		
		if(pwm < 10) //if too small
		{
			pwm = 10;
		}
		if(pwm > 500) //if too big
		{
			pwm = 500;
		}
		else
	
		OCR1A = pwm; //regulate!
	}
	
	if(shutdownflag == 1) //ramp it down!
	{
		if(OCR1A >= 30)
		{
			OCR1A = OCR1A - 3;
		}
		else
		{
			shutdownflag = 2;
		}
	}
}

ISR(TIMER2_COMPA_vect) //used for the rampup
{
	if(cntrlstate == 0 && startupflag == 0)
	{
		if(OCR1A <= 130) //ramp it up!
		{
			OCR1A = OCR1A + 1;
		}
		else //rampup finished ==> let the controller act!
		{
			TIMSK2 &= ~(1<<OCIE2A); //disable Compare-Interrupt 2
			cntrlstate = 1; //let the controller act.
			commcnt = 0; //reset commutation counter to the beginning value
			accu = 150;
			TCNT2 = 0;
		}
	}
}

ISR(ADC_vect) //ADC conversion complete
{
	adccnt++;
	adcvalue = adcvalue + ADC;
}

ISR(WDT_vect )
{
	wdtovf = 1;
	boostcnt++;
}
//#########################