/*
These are the functions used for the startup of the BLDC-Motor. 
Mario Mauerer 2011
*/

#include <avr/io.h>    
#include <util/delay.h>

#include "main_saw.h" //needed for the global variables
#include "uart_saw.h" //needed for the messages

//Perform a self-check of the system prior to start: Measure DC-Link, Current and Temperature. If everything is OK: Start the saw.
void system_check(void)
{
	uint16_t value = 0; //variable for random stuff
	
//Display some Information over the UART:
	uart_stringsend("Hello");
	uart_stringsend("This is the BASTLI circular saw");
	uart_stringsend("Built by:");
	uart_stringsend("Mario Mauerer");
	uart_stringsend("and");
	uart_stringsend("Lukas Schrittwieser");
	uart_stringsend("April 2011");
	uart_stringsend(" ");
	uart_stringsend("The saw is now starting!");
	uart_stringsend("performing self-check...");
	
	_delay_ms(400); //make a pause to charge all capacitors
	
//Measure DC-Link-Voltage:
	//DC-Link-Voltage = 0.045 * ADC + 0.047
	//configure ADC for Port ADC1 : DC-Link-Voltage:
	ADCSRA &= ~(1<<ADSC); //ADC Disable in order to change channels.
	ADMUX &= 0b11110000; //Clear current Channel Selection
	ADMUX |= (1<<MUX0); //ADC1 as Input (DC-Link-Voltage)
	ADCSRA |= (1<<ADSC); //restart ADC
	ADCSRA |= (1<<ADIE); //enable the conversion complete Interrupt
	adccnt = 0;
	adcvalue = 0;
	while(adccnt < 16) //take 16 measurements (using the ADC-Interrupt), wait until ADC has finished
	{
	}
	ADCSRA &= ~(1<<ADIE); //disable the conversion complete Interrupt
	value = adcvalue >> 4; //divide by 16 to calculate average
	
	if(value < 700 || value > 843) //if DC-Link-Voltage larger than 38V or smaller than 33V (730)==> not good for starting ==> error 
	{
		value = (45*value)+47;
		value = value / 1000; //calculate the voltage (rough calculation), don't use floats.
		
		uart_stringsend("DC link error!");
		_delay_ms(5);
		uart_stringsend("DC-link voltage:");
		_delay_ms(5);
		uart_intsend(value);
		
		while(1) //let the LED blink TWO times when having a DC-Link-Error and wait forever here. 
		{
			LED_on;
			_delay_ms(100);
			LED_off;
			_delay_ms(100);
			LED_on;
			_delay_ms(100);
			LED_off;
			_delay_ms(500);
		}
	}
	
	
//Measure Current: Should be Zero:
	//Current = 0.0181 * ADC + 1.88
	//configure ADC for Port ADC0: Current
	ADCSRA &= ~(1<<ADSC); //ADC Disable in order to change channels.
	ADMUX &= 0b11110000; //ADC0 as Input (Current-Measurement)
	ADCSRA |= (1<<ADSC); //restart ADC
	ADCSRA |= (1<<ADIE); //enable the conversion complete Interrupt
	adccnt = 0;
	adcvalue = 0;
	while(adccnt < 16) //take 16 measurements (using the ADC-Interrupt)
	{
	}
	ADCSRA &= ~(1<<ADIE); //disable the conversion complete Interrupt
	value = adcvalue >> 4; //divide by 16 to build average
	
	if(value > 70) //some current is flowing ==> not good at startup, there should be no current at all ==> error
	{
		
		value = (18*value)+1880;
		value = value / 1000; //calculate the voltage (rough calculation), don't use floats.
		
		uart_stringsend("Current error!");
		_delay_ms(5);
		uart_stringsend("Current (A):");
		_delay_ms(5);
		uart_intsend(value);
		while(1) //let the LED blink FOUR times when having a current-error and wait here forever
		{
			LED_on;
			_delay_ms(200);
			LED_off;
			_delay_ms(200);
			LED_on;
			_delay_ms(200);
			LED_off;
			_delay_ms(200);
			LED_on;
			_delay_ms(200);
			LED_off;
			_delay_ms(200);
			LED_on;
			_delay_ms(200);
			LED_off;
			_delay_ms(600);
		}
	}
	
//Measure Temperature of the Motor: Should be OK:
	//Temperature (C) = 0.442 * ADC - 188
	//configure ADC for Port ADC2 (Temp-Sensor)
	ADCSRA &= ~(1<<ADSC); //ADC Disable in order to change channels.
	ADMUX &= 0b11110000; //Clear current Channel Selection
	ADMUX |= (1<<MUX1); //ADC2 as Input (Temp-Sensor)
	ADCSRA |= (1<<ADSC); //restart ADC
	ADCSRA |= (1<<ADIE); //enable the conversion complete Interrupt
	adccnt = 0;
	adcvalue = 0;
	while(adccnt < 16) //take 16 measurements (using the ADC-Interrupt)
	{
	}
	ADCSRA &= ~(1<<ADIE); //disable the conversion complete Interrupt
	value = adcvalue >> 4; //divide by 16 to build average
	
	if(value > 545) //if temp > 53C ==> too hot to start ==> error
	{
		uint16_t temp = 0;
		
		temp = 44*value - 18800;
		temp = temp / 100;
		value = value * 2;
		value = value / 1000;
		value = value + temp; //calculate the voltage (rough calculation), don't use floats.
		
		uart_stringsend("Temperature error!");
		_delay_ms(5);
		uart_stringsend("Temperature (C):");
		_delay_ms(5); 
		uart_intsend(value);
		
		while(1) //let the LED blink THREE times when having temperature-error and wait forever here.
		{
			LED_on;
			_delay_ms(200);
			LED_off;
			_delay_ms(200);
			LED_on;
			_delay_ms(200);
			LED_off;
			_delay_ms(200);
			LED_on;
			_delay_ms(200);
			LED_off;
			_delay_ms(600);
		}
	}
	
	//else: Everything is fine and the motor may start!
	uart_stringsend("self-check OK");
	uart_stringsend("The saw starts now!");

}

//This routine brings the motor to its starting-position and beeps three times using the motor itself. 
void motor_beep(void)
{
	for(uint8_t i = 0; i<3;i++) //make three beeps with the motor.
	{
		CLOW_on; //in order for bootstrap-C to charge
		_delay_ms(50); //charge bootstrap-C
		CLOW_off;
		//Starting-Position: CHIGH_on; ALOW_on; BLOW_on.
		
		for(uint8_t j = 0; j<200; j++)
		{
			AHIGH_off;
			BHIGH_off;
			
			CLOW_off;
			CHIGH_on;
			
			BLOW_on;
			ALOW_on;
		
			_delay_us(170);
			
			CHIGH_off;
			CLOW_on;
			
			BLOW_off;
			ALOW_off;
			
			AHIGH_on;
			BHIGH_on;
			
			_delay_us(170);
		}
		
			AHIGH_off;
			BHIGH_off;
			CHIGH_off;
			BLOW_off;
			ALOW_off;
			CLOW_off;
		
		_delay_ms(160);
	}
	CHIGH_on;
	BLOW_on;
	ALOW_on;
	_delay_ms(10);
	AHIGH_off;
	BHIGH_off;
	CHIGH_off;
	BLOW_off;
	ALOW_off;
	CLOW_off;
	
	_delay_ms(50);
}